package filegateway

import (
	"context"
	"fmt"
	"testing"
	"time"

	nrpc "github.com/nats-rpc/nrpc"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"orus.io/xbus/xbus-stdlib/service"
	"orus.io/xbus/xbus-stdlib/service/svctest"
	"xbus.io/go-xbus/v4/envelope"
	"xbus.io/go-xbus/v4/fullenv"
	"xbus.io/go-xbus/v4/testutils"
)

func makeChunkedEnvelope(t *testing.T, msgType string, chunks []string) envelope.Envelope {
	t.Helper()
	msg, w := envelope.MustNewMessageChunkWriter(msgType)
	for _, c := range chunks {
		require.NoError(t, w.Write([]byte(c)))
	}
	require.NoError(t, w.Close())
	return envelope.MustNewEnvelope(msg)
}

func TestConsumer(t *testing.T) {
	t.Run("basics", func(t *testing.T) {
		tester := svctest.TestConsumer(t, NewConsumer, nil,
			service.Settings{
				"target": "mock://testconsumer/targetdir/",
			},
		)

		require.NoError(t, tester.Run(svctest.Inputs{
			"default": makeChunkedEnvelope(t, "test", []string{
				`{"filename":"file.txt"}`,
				"Hello World!",
			}),
		}))

		assert.Equal(t, "Hello World!", string(getMemoryFilestem("testconsumer").Files["/targetdir/file.txt"]))
	})
	t.Run("temporary errors returns SERVERTOOBUSY", func(t *testing.T) {
		tester := svctest.TestConsumer(t, NewConsumer, nil,
			service.Settings{
				"target":           "mock://testconsumer/targetdir/",
				"target.fserror":   "onwrite",
				"temporary-errors": true,
			})

		tester.Logger.ExpectsError(testutils.ExpectsContains("SERVERTOOBUSY"))

		err := tester.Run(svctest.Inputs{
			"default": makeChunkedEnvelope(t, "test", []string{
				`{"filename":"file.txt"}`,
				"Hello World!",
			}),
		})

		if nErr, ok := err.(*nrpc.Error); assert.True(t, ok, "Error is: %s", err) {
			assert.Equal(t, nrpc.Error_SERVERTOOBUSY, nErr.Type)
			assert.Equal(t, "[filegateway] Error opening file '/targetdir/file.txt': write error.", nErr.Message)
		}
	})
	t.Run("fullenv temporary errors", func(t *testing.T) {
		log := testutils.GetTestLogger(t)
		log.ExpectsError(testutils.ExpectsContains("SERVERTOOBUSY"))
		log.ExpectsError(testutils.ExpectsContains("SERVERTOOBUSY"))
		log.ExpectsError(testutils.ExpectsContains("SERVERTOOBUSY"))
		env := fullenv.NewFullenv("consumer_test.yaml", log.Logger)
		require.NoError(t, env.Start())
		defer func() { _ = env.Stop() }()

		client, err := env.LoadClient("client")
		if err != nil {
			t.Fatal(err)
		}
		if err := client.Startup(true); err != nil {
			t.Fatal(err)
		}
		defer client.Shutdown()

		emitter := client.GetActorByName("emitter")
		require.NotNil(t, emitter, "No 'emitter'")

		msg, w, err := envelope.NewMessageChunkWriter("msg")
		if err != nil {
			t.Fatal(err)
		}
		if !assert.NoError(t, w.WriteString(`{"filename": "test.txt"}`)) ||
			!assert.NoError(t, w.WriteString(`SomeContent`)) ||
			!assert.NoError(t, w.Close()) {
			t.FailNow()
		}
		ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
		defer cancel()
		emit, err := emitter.Emit(ctx, envelope.MustNewEnvelope(msg))
		if !assert.NoError(t, err) {
			t.FailNow()
		}
		completed := make(chan error)
		go func() {
			err := emit.WaitCompletion()
			t.Log("WaitCompletion is over, result is", err)
			completed <- err
		}()
		fmt.Println("failed 1", t.Failed())
		// Now the process should remain 'RUNNING' and never end because the
		// consumer returns SERVERTOOBUSY
		select {
		case <-ctx.Done():
			// timeout reached, as expected
			t.Log("Timeout reached as expected")
		case err := <-completed:
			if ctx.Err() != nil && err == ctx.Err() {
				t.Log("Timeout reached as expected (indirect)")
			} else {
				t.Error("The process terminated with", err)
			}
		}
		fmt.Println("final failed", t.Failed())
	})
}
