package filegateway

import (
	"bytes"
	"context"
	"encoding/json"
	"fmt"
	"io"
	"path"
	"path/filepath"
	"strings"
	"sync"
	"text/template"

	stdlib "orus.io/xbus/xbus-stdlib/lib"
	xbus "xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/envelope"
)

// RenameTargetVars holds the variables available when rendering a rename target
// path template
type RenameTargetVars struct {
	Full       string
	Dir        string
	Base       string
	Ext        string
	WithoutExt string

	EnvelopeID string
}

type EmitEnvelopeFunc func(
	ctx context.Context,
	env envelope.Envelope,
) error

// Emitter watches a pattern, emit the matching files into the bus, and delete
// the file after emission
type Emitter struct {
	emitEnvelope     EmitEnvelopeFunc
	log              xbus.Logger
	fs               Filesystem
	pattern          string
	msgType          string
	postAction       string
	renameTemplate   *template.Template
	smallFileSize    int
	streamBufferSize int

	exit       chan struct{}
	paths      chan string
	observeErr chan error
	wg         sync.WaitGroup

	smallFileBuffer  []byte
	streamBufferList [][]byte
}

// NewEmitter creates a Emitter
func NewEmitter(actor *xbus.Actor, settings xbus.Settings) (xbus.ActorService, error) {
	log := actor.Log
	options, err := LoadFilesystemOptions(settings, "source")
	if err != nil {
		return nil, fmt.Errorf("Could not read 'source' setting: %w", err)
	}
	fs, err := GetFilesystem(options.Scheme, options, log.AsZero())
	if err != nil {
		return nil, fmt.Errorf("Could not get 'source' filesystem: %w", err)
	}
	emitter := Emitter{
		emitEnvelope: func(ctx context.Context, env envelope.Envelope) error {
			_, err := actor.Emit(ctx, env)
			return err
		},
		log:              actor.Log,
		fs:               fs,
		pattern:          options.Path,
		msgType:          settings.MustString("message-type"),
		smallFileSize:    settings.MustIntD("small-file-size", 10),
		streamBufferSize: settings.MustIntD("stream-size", 10),
	}
	switch postAction := settings.MustStringD("post-action", "delete"); {
	case postAction == "delete":
		emitter.postAction = "delete"
	case strings.HasPrefix(postAction, "rename "):
		emitter.postAction = "rename"
		t, err := template.New("rename target").Parse(postAction[7:])
		if err != nil {
			return nil, fmt.Errorf("post-action: error parsing rename target: %w", err)
		}
		emitter.renameTemplate = t
	default:
		log.Err(err).Msg("Invalid 'post-action'")
		return nil, fmt.Errorf("Invalid 'post-action': %w", err)
	}
	return &emitter, nil
}

// Startup lauch the observe routine
func (e *Emitter) Startup() error {
	e.exit = make(chan struct{})
	e.paths = make(chan string)
	e.observeErr = make(chan error, 1)

	if e.smallFileSize == 0 {
		e.smallFileSize = 10
	}
	e.smallFileBuffer = make([]byte, e.smallFileSize)

	if e.streamBufferSize < 3 {
		e.streamBufferSize = 3
	}
	e.streamBufferList = make([][]byte, e.streamBufferSize)
	for i := range e.streamBufferList {
		e.streamBufferList[i] = make([]byte, 1024*1024)
	}

	ctx, cancel := context.WithCancel(context.Background())

	e.wg.Add(1)
	go func() {
		defer e.wg.Done()

		err := e.fs.Observe(ctx, e.pattern, e.paths, e.log)
		if err != nil {
			cancel()
			e.observeErr <- err
		}
	}()

	select {
	case err := <-e.observeErr:
		return err
	case <-e.paths:
		// the observer is ready
	}

	e.wg.Add(1)
	go e.run(ctx, cancel)

	return nil
}

// Shutdown stops the observe routine
func (e *Emitter) Shutdown() error {
	close(e.exit)
	e.wg.Wait()
	return nil
}

func (e *Emitter) run(ctx context.Context, ctxCancel func()) {
	defer e.wg.Done()
	defer ctxCancel()

	for {
		select {
		case fname := <-e.paths:
			e.log.Debug().Str("filename", fname).Msg("Emitting file in the bus")
			if err := e.emit(ctx, fname); err != nil {
				e.log.Err(err).Str("filename", fname).Msg("Could not emit file")
			}

		case <-e.exit:
			return
		case <-ctx.Done():
			return
		}
	}
}

func (e *Emitter) doPostAction(fpath string, envelopeID api.UUID) error {
	switch e.postAction {
	case "delete":
		return e.fs.Delete(fpath)
	case "rename":
		buf := bytes.NewBuffer(nil)
		base := filepath.Base(fpath)
		if err := e.renameTemplate.Execute(buf, RenameTargetVars{
			Full:       fpath,
			Dir:        filepath.Dir(fpath),
			Base:       base,
			Ext:        filepath.Ext(fpath),
			WithoutExt: base[:len(base)-len(filepath.Ext(base))],
			EnvelopeID: envelopeID.String(),
		}); err != nil {
			return err
		}
		return e.fs.Rename(fpath, buf.String())
	}
	return fmt.Errorf("invalid post-action: %s", e.postAction)
}

func (e *Emitter) emit(ctx context.Context, fpath string) error {
	log := e.log.AsZero().With().Str("filepath", fpath).Logger()

	meta := stdlib.FileMeta{
		Filename: path.Base(fpath),
	}
	metaHeader, err := json.Marshal(meta)
	if err != nil {
		return err
	}

	reader, err := e.fs.Read(fpath)
	if err != nil {
		return err
	}
	defer func() {
		if err := reader.Close(); err != nil {
			log.Err(err).Msg("failed to close remote file")
		}
	}()
	var (
		curBuffer     = 0
		curBufferSize int

		readComplete = false
	)

	// Read up to smallFileSize mb before sending anything to the bus
	for {
		n, err := reader.Read(e.smallFileBuffer[curBufferSize:])
		curBufferSize += n
		if err != nil && err != io.EOF {
			return fmt.Errorf("could not read %s: %w", fpath, err)
		}
		if err == io.EOF {
			readComplete = true
			break
		}
		if curBufferSize == len(e.smallFileBuffer) {
			break
		}
	}

	msgWriter := envelope.MustNewMessageWriter(e.msgType)
	env := envelope.MustNewEnvelope(msgWriter)

	log = log.With().Str("envelope_id", env.ID().String()).Logger()

	if err := msgWriter.WriteChunk(metaHeader); err != nil {
		log.Err(err).Msg("failed to write metadata header in the message")
		return err
	}

	if _, err := msgWriter.Write(e.smallFileBuffer[:curBufferSize]); err != nil {
		log.Err(err).Msg("failed to write metadata header in the message")
		return err
	}

	var needStreamedEmission = !readComplete

	if !readComplete {
		curBufferSize, err = reader.Read(e.streamBufferList[curBuffer])
		if curBufferSize == 0 && err == io.EOF {
			needStreamedEmission = false
			readComplete = true
		}
	}

	if readComplete {
		if err := msgWriter.Close(); err != nil {
			log.Err(err).Msg("failed to close the message")
			return err
		}
	}

	if needStreamedEmission {
		var wg sync.WaitGroup
		defer wg.Wait()
		dataC := make(chan []byte, e.streamBufferSize-2)
		done := make(chan struct{})

		wg.Add(1)
		go func() {
			defer close(done)
			defer wg.Done()
			for data := range dataC {
				if _, err := msgWriter.Write(data); err != nil {
					log.Err(err).Msg("failed to write data in the message")
					return
				}
			}
			if err := msgWriter.Close(); err != nil {
				log.Err(err).Msg("failed to close the message")
			}
		}()

		wg.Add(1)
		go func() {
			defer wg.Done()
			for {
				select {
				case <-done:
					return
				default:
				}
				select {
				case dataC <- e.streamBufferList[curBuffer][:curBufferSize]:
					curBuffer = (curBuffer + 1) % e.streamBufferSize
				case <-done:
					return
				}

				curBufferSize, err = reader.Read(e.streamBufferList[curBuffer])
				if err == io.EOF {
					close(dataC)
					return
				}
				if err != nil {
					close(dataC)
					closeWithError := fmt.Errorf("could not read input file: %w", err)

					if closeErr := env.CloseWithError(closeWithError); closeErr != nil {
						log.Err(closeErr).Msg("failed to close envelope with error")
					}
					return
				}
			}
		}()
	}

	// Now the envelope is ready, or being filled by the above routines
	if err := e.emitEnvelope(ctx, env); err != nil {
		return err
	}

	log.Info().
		Msg("Emitted file")
	return e.doPostAction(fpath, env.ID())
}

func init() {
	xbus.RegisterActorService("file-gateway-emitter", NewEmitter)
}
