package filegateway

import (
	"bytes"
	"context"
	"encoding/binary"
	"errors"
	"io"
	"os"
	"strings"
	"sync"
	"testing"
	"time"

	"github.com/rs/zerolog"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/testutils"
)

func writeFile(fs Filesystem, path, content string) error {
	return writeFileBytes(fs, path, []byte(content))
}

func writeFileBytes(fs Filesystem, path string, content []byte) error {
	writer, err := fs.Write(path, -1)
	if err != nil {
		return err
	}
	if _, err := writer.Write(content); err != nil {
		return err
	}
	return writer.Close()
}

func FilesystemTestSuite(t *testing.T, factory FilesystemFactory, options FilesystemOptions) {
	options.PollingInterval = 10 * time.Millisecond
	t.Run("Write and read", func(t *testing.T) {
		fs, err := factory(options, *testutils.GetTestLogger(t).Zero())
		if err != nil {
			t.Fatal(err)
		}
		writer, err := fs.Write(t.Name(), -1)
		if err != nil {
			t.Fatal(err)
		}
		if _, err := writer.Write([]byte("Hello World!")); err != nil {
			t.Fatal(err)
		}
		if err := writer.Close(); err != nil {
			t.Fatal(err)
		}
		reader, err := fs.Read(t.Name())
		if err != nil {
			t.Fatal(err)
		}
		defer func() {
			assert.NoError(t, reader.Close())
		}()
		b, err := io.ReadAll(reader)
		if err != nil {
			t.Fatal(err)
		}
		assert.Equal(t, "Hello World!", string(b))
	})

	t.Run("Big file", func(t *testing.T) {
		if strings.Contains(t.Name(), "Vsftpd") {
			t.Skip("Vsftpd container has issues with big files, skipping")
		}
		fs, err := factory(options, *testutils.GetTestLogger(t).Zero())
		require.NoError(t, err)

		fname := t.Name()

		count := 50000

		t.Run("Write", func(t *testing.T) {
			writer, err := fs.Write(fname, -1)
			require.NoError(t, err)
			defer func() {
				assert.NoError(t, writer.Close())
			}()

			var buf = make([]byte, 8)
			for i := 0; i != count; i++ {
				binary.LittleEndian.PutUint64(buf, uint64(i))
				v := binary.LittleEndian.Uint64(buf)
				require.Equal(t, uint64(i), v)

				c, err := writer.Write(buf)
				require.NoError(t, err)
				require.Equal(t, 8, c)
			}
		})

		t.Run("Read", func(t *testing.T) {
			reader, err := fs.Read(fname)
			require.NoError(t, err)
			defer func() {
				assert.NoError(t, reader.Close())
			}()

			var buf = make([]byte, 8)
			for i := 0; i != count; i++ {
				c, err := io.ReadFull(reader, buf)
				require.NoError(t, err)
				assert.Equal(t, 8, c)
				v := binary.LittleEndian.Uint64(buf)
				require.Equal(t, i, int(v))
			}
		})
	})

	t.Run("Delete", func(t *testing.T) {
		fs, err := factory(options, *testutils.GetTestLogger(t).Zero())
		if err != nil {
			t.Fatal(err)
		}
		if err := fs.Delete(t.Name()); !os.IsNotExist(err) {
			t.Fatal("Expected os.ErrNotExist, got", err)
		}
		writer, err := fs.Write(t.Name(), -1)
		if err != nil {
			t.Fatal(err)
		}
		if _, err := writer.Write([]byte("Hello World!")); err != nil {
			t.Fatal(err)
		}
		if err := writer.Close(); err != nil {
			t.Fatal(err)
		}
		if err := fs.Delete(t.Name()); err != nil {
			t.Fatal(err)
		}
		_, err = fs.Read(t.Name())
		if err == nil {
			t.Fatal("Expects a ErrNotExist, got nil")
		}
		assert.Equal(t, true, os.IsNotExist(err), err.Error())
	})

	t.Run("Rename", func(t *testing.T) {
		fs, err := factory(options, *testutils.GetTestLogger(t).Zero())
		if err != nil {
			t.Fatal(err)
		}
		if err := fs.Rename(t.Name(), t.Name()+".1"); !os.IsNotExist(err) {
			t.Fatal("Expected os.ErrNotExist, got", err)
		}
		writer, err := fs.Write(t.Name(), -1)
		if err != nil {
			t.Fatal(err)
		}
		if _, err := writer.Write([]byte("Hello World!")); err != nil {
			t.Fatal(err)
		}
		if err := writer.Close(); err != nil {
			t.Fatal(err)
		}
		if err := fs.Rename(t.Name(), t.Name()+".1"); err != nil {
			t.Fatal(err)
		}
		reader, err := fs.Read(t.Name() + ".1")
		if err != nil {
			t.Fatal(err)
		}
		defer reader.Close()
		b, err := io.ReadAll(reader)
		if err != nil {
			t.Fatal(err)
		}
		assert.Equal(t, "Hello World!", string(b))
	})

	t.Run("WaitForDeletion", func(t *testing.T) {
		t.Run("Returns when the file is not here", func(t *testing.T) {
			fs, err := factory(options, *testutils.GetTestLogger(t).Zero())
			if err != nil {
				t.Fatal(err)
			}
			ctx, cancel := context.WithTimeout(context.Background(), time.Second)
			defer cancel()
			assert.NoError(t, fs.WaitForDeletion(ctx, t.Name()))
		})
		t.Run("Is cancelable", func(t *testing.T) {
			fs, err := factory(options, *testutils.GetTestLogger(t).Zero())
			if err != nil {
				t.Fatal(err)
			}
			writer, err := fs.Write(t.Name(), -1)
			if err != nil {
				t.Fatal(err)
			}
			if _, err := writer.Write([]byte("Hello World!")); err != nil {
				t.Fatal(err)
			}
			if err := writer.Close(); err != nil {
				t.Fatal(err)
			}
			ctx, f := context.WithTimeout(context.Background(), 50*time.Millisecond)
			defer f()
			err = fs.WaitForDeletion(ctx, t.Name())
			if err == nil {
				t.Fatal("Should receive a DeadlineExceeded error, got nil")
			}
			assert.Equal(t, context.DeadlineExceeded, err, err.Error())
		})
		t.Run("Returns when the file is deleted", func(t *testing.T) {
			fs, err := factory(options, *testutils.GetTestLogger(t).Zero())
			if err != nil {
				t.Fatal(err)
			}
			ctx, cancel := context.WithTimeout(context.Background(), time.Second)
			defer cancel()
			assert.NoError(t, fs.WaitForDeletion(ctx, t.Name()))
		})
	})

	t.Run("Observe", func(t *testing.T) {
		t.Run("Is cancelable", func(t *testing.T) {
			logger := testutils.GetTestLogger(t)
			fs, err := factory(options, *testutils.GetTestLogger(t).Zero())
			if err != nil {
				t.Fatal(err)
			}
			var pattern = t.Name() + "/*.txt"

			done := make(chan struct{})
			defer func() {
				select {
				case <-done:
				case <-time.After(5 * time.Second):
					t.Error("Observe never ended")
				}
			}()

			ctx, cancel := context.WithCancel(context.Background())
			defer cancel()

			paths := make(chan string)

			go func() {
				defer close(done)
				defer close(paths)
				if err := fs.Observe(ctx, pattern, paths, logger.XbusLogger); err != nil {
					if !errors.Is(err, context.Canceled) {
						assert.NoError(t, err)
					}
				}
			}()
		})
		t.Run("Is cancelable when chan is full", func(t *testing.T) {
			logger := testutils.GetTestLogger(t)
			fs, err := factory(options, *testutils.GetTestLogger(t).Zero())
			if err != nil {
				t.Fatal(err)
			}
			fname1 := t.Name() + "/f1.txt"
			fname2 := t.Name() + "/f2.txt"
			assert.NoError(t, writeFile(fs, fname1, "f1 content"))
			assert.NoError(t, writeFile(fs, fname2, "f2 content"))

			var pattern = t.Name() + "/*.txt"

			done := make(chan struct{})
			defer func() {
				select {
				case <-done:
				case <-time.After(5 * time.Second):
					t.Error("Observe never ended")
				}
			}()

			ctx, cancel := context.WithCancel(context.Background())
			defer cancel()

			paths := make(chan string)

			go func() {
				defer close(done)
				defer close(paths)
				if err := fs.Observe(ctx, pattern, paths, logger.XbusLogger); err != nil {
					if !errors.Is(err, context.Canceled) {
						assert.NoError(t, err)
					}
				}
			}()

			select {
			case <-ctx.Done():
				require.NoError(t, ctx.Err())
			case p := <-paths:
				require.Equal(t, "", p)
			}

			cancel()
		})
		t.Run("Returns the files that matches already", func(t *testing.T) {
			logger := testutils.GetTestLogger(t)
			fs, err := factory(options, *testutils.GetTestLogger(t).Zero())
			if err != nil {
				t.Fatal(err)
			}
			fname1 := t.Name() + "/f1.txt"
			fname2 := t.Name() + "/f2.txt"
			fname3 := t.Name() + "/f3.log"
			assert.NoError(t, writeFile(fs, fname1, "f1 content"))
			assert.NoError(t, writeFile(fs, fname2, "f2 content"))
			assert.NoError(t, writeFile(fs, fname3, "f3 content"))

			var pattern = t.Name() + "/*.txt"

			done := make(chan struct{})
			defer func() {
				select {
				case <-done:
				case <-time.After(5 * time.Second):
					t.Error("Observe never ended")
				}
			}()

			ctx, cancel := context.WithTimeout(context.Background(), 10*time.Second)
			defer cancel()

			paths := make(chan string)

			go func() {
				defer close(done)
				defer close(paths)
				if err := fs.Observe(ctx, pattern, paths, logger.XbusLogger); err != nil {
					if !errors.Is(err, context.Canceled) {
						assert.NoError(t, err)
					}
				}
			}()

			select {
			case fname := <-paths:
				assert.Equal(t, "", fname)
			case <-ctx.Done():
				t.Error("never received observe 'ready'")
				return
			}

			select {
			case fname := <-paths:
				assert.Equal(t, fname1, fname)
			case <-ctx.Done():
				t.Error("fname1 was never sent")
				return
			}

			select {
			case fname := <-paths:
				assert.Equal(t, fname2, fname)
			case <-ctx.Done():
				t.Error("fname2 was never sent")
				return
			}

			cancel()
		})
		t.Run("Returns new files when they arrive", func(t *testing.T) {
			logger := testutils.GetTestLogger(t)
			fs, err := factory(options, *testutils.GetTestLogger(t).Zero())
			if err != nil {
				t.Fatal(err)
			}
			fname1 := t.Name() + "/f1.txt"
			fname2 := t.Name() + "/f2.txt"
			var pattern = t.Name() + "/*.txt"

			done := make(chan string)
			defer func() {
				var doneRoutines []string
				timeout := time.After(5 * time.Second)
				for len(doneRoutines) < 2 {
					select {
					case s := <-done:
						doneRoutines = append(doneRoutines, s)
					case <-timeout:
						t.Error("some routines did not finish. The finished ones are", doneRoutines)
						return
					}
				}
			}()

			ctx, cancel := context.WithTimeout(context.Background(), 10*time.Second)
			defer cancel()

			paths := make(chan string)
			go func() {
				defer func() { done <- "observe" }()
				defer close(paths)

				if err := fs.Observe(ctx, pattern, paths, logger.XbusLogger); err != nil {
					if !errors.Is(err, context.Canceled) {
						assert.NoError(t, err)
					}
				}
			}()

			select {
			case fname := <-paths:
				assert.Equal(t, "", fname)
			case <-ctx.Done():
				t.Error("never received observe 'ready'")
				return
			}

			go func() {
				defer func() { done <- "write" }()
				select {
				case <-ctx.Done():
					return
				case <-time.After(30 * time.Millisecond):
					if err := writeFile(fs, fname1, "f1 content"); err != nil {
						t.Error(err)
					}
				}
				select {
				case <-ctx.Done():
					return
				case <-time.After(50 * time.Millisecond):
					if err := writeFile(fs, fname2, "f2 content"); err != nil {
						t.Error(err)
					}
				}
			}()

			select {
			case fname := <-paths:
				assert.Equal(t, fname1, fname)
			case <-ctx.Done():
				t.Error("timeout waiting for fname1")
				return
			}
			select {
			case fname := <-paths:
				assert.Equal(t, fname2, fname)
			case <-ctx.Done():
				t.Error("timeout waiting for fname2")
				return
			}
		})
	})
}

func TestMemoryFilesystem(t *testing.T) {
	FilesystemTestSuite(t, memoryFilesystemFactory, FilesystemOptions{Host: "."})
}

// Below is a memory implementation of the Filesystem interface, for testing
// purpose

var memfs = make(map[string]*MemoryFilesystem)

type MemoryFilesystem struct {
	host            string
	Files           map[string][]byte
	pollingInterval time.Duration
	fsWriteError    bool
	m               sync.Mutex
}

func NewMemoryFilesystem(host string) *MemoryFilesystem {
	fs := MemoryFilesystem{
		host:            host,
		Files:           make(map[string][]byte),
		pollingInterval: time.Second,
	}
	return &fs
}

func (fs *MemoryFilesystem) String() string {
	return "mock://" + fs.host
}

type memFileWriter struct {
	*bytes.Buffer

	fs   *MemoryFilesystem
	path string
}

func (w *memFileWriter) Close() error {
	w.fs.m.Lock()
	defer w.fs.m.Unlock()

	w.fs.Files[w.path] = w.Buffer.Bytes()
	return nil
}

var ErrWriteError = errors.New("write error")

func (fs *MemoryFilesystem) Write(path string, size int) (io.WriteCloser, error) {
	if fs.fsWriteError {
		return nil, ErrWriteError
	}
	return &memFileWriter{
		Buffer: bytes.NewBuffer(nil),
		fs:     fs,
		path:   path,
	}, nil
}

func (fs *MemoryFilesystem) list() (filepaths []string) {
	fs.m.Lock()
	defer fs.m.Unlock()
	for p := range fs.Files {
		filepaths = append(filepaths, p)
	}
	return
}

func (fs *MemoryFilesystem) Observe(
	ctx context.Context, pattern string, paths chan<- string, logger xbus.Logger,
) error {
	fo := NewFileObserver(ctx, pattern, "", paths)
	select {
	case paths <- "":
	case <-ctx.Done():
		return ctx.Err()
	}

	for {
		if err := fo.AddFiles(fs.list()...); err != nil {
			return err
		}
		select {
		case <-time.After(fs.pollingInterval):
		case <-ctx.Done():
			return ctx.Err()
		}
	}
}

type FileReader struct {
	*bytes.Buffer
}

func (fr FileReader) Close() error {
	return nil
}

func (fs *MemoryFilesystem) Read(path string) (io.ReadCloser, error) {
	fs.m.Lock()
	defer fs.m.Unlock()

	if buf, ok := fs.Files[path]; ok {
		return FileReader{bytes.NewBuffer(buf)}, nil
	}
	return nil, os.ErrNotExist
}

func (fs *MemoryFilesystem) WaitForDeletion(ctx context.Context, path string) error {
	for {
		fs.m.Lock()
		_, ok := fs.Files[path]
		fs.m.Unlock()
		if !ok {
			return nil
		}
		select {
		case <-ctx.Done():
			return ctx.Err()
		case <-time.After(fs.pollingInterval):
		}
	}
}

func (fs *MemoryFilesystem) Rename(path, newpath string) error {
	fs.m.Lock()
	defer fs.m.Unlock()

	if buf, ok := fs.Files[path]; ok {
		delete(fs.Files, path)
		fs.Files[newpath] = buf
		return nil
	}
	return os.ErrNotExist
}

func (fs *MemoryFilesystem) Delete(path string) error {
	fs.m.Lock()
	defer fs.m.Unlock()

	if _, ok := fs.Files[path]; ok {
		delete(fs.Files, path)
		return nil
	}
	return os.ErrNotExist
}

func getMemoryFilestem(host string) *MemoryFilesystem {
	if fs, ok := memfs[host]; ok {
		return fs
	}
	fs := NewMemoryFilesystem(host)
	memfs[host] = fs
	return fs
}

func memoryFilesystemFactory(options FilesystemOptions, log zerolog.Logger) (Filesystem, error) {
	fs := getMemoryFilestem(options.Host)
	fs.pollingInterval = options.PollingInterval
	fserror, err := options.GetCustomStringD("fserror", "")
	if err != nil {
		return nil, err
	}
	fs.fsWriteError = fserror == "onwrite"
	return fs, nil
}

func init() {
	MustRegisterFilesystem("mock", memoryFilesystemFactory)
}
