package routing

import (
	"context"
	"fmt"

	libjq_go "github.com/flant/libjq-go"
	"github.com/flant/libjq-go/pkg/jq"
	stdlib "orus.io/xbus/xbus-stdlib/lib"
	"orus.io/xbus/xbus-stdlib/service"

	"xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"
)

// NewImperativeRouter creates a ImperativeRouter
func NewImperativeRouter(actor *api.Actor, settings service.Settings) (*ImperativeRouter, error) {
	outputJQ, err := settings.String("extract-output-jq")
	if err != nil {
		return nil, err
	}
	return &ImperativeRouter{
		outputJQ: outputJQ,
	}, nil
}

// NewImperativeRouterAsActorService creates a ImperativeRouter and returns it
// as a service.Worker
func NewImperativeRouterAsActorService(
	actor *api.Actor, settings service.Settings, log xbus.Logger,
) (service.Worker, error) {
	return NewImperativeRouter(actor, settings)
}

// ImperativeRouter is a Xbus Router Service that routes an incoming event
// to the output designated by some data field inside the event itself.
// It currently has one setting:
// - 'output-jq': a jq expression that produces the output name, assuming the payload
//   is encoded in json
type ImperativeRouter struct {
	outputJQ        string
	outputJQProgram *jq.JqProgram
}

// Startup precompiles the jq program
func (router *ImperativeRouter) Startup() error {
	program, err := libjq_go.Jq().Program(router.outputJQ).Precompile()
	if err != nil {
		return err
	}
	router.outputJQProgram = program
	return nil
}

// Process handles a Actor ProcessRequest
func (router *ImperativeRouter) Process(
	ctx context.Context, r *api.ActorProcessRequest,
	inputs service.Inputs, outputs service.Outputs, log xbus.Logger,
) error {
	if router.outputJQProgram == nil {
		return fmt.Errorf("not initialised")
	}
	env, err := inputs.Get(api.DefaultInput)
	if err != nil {
		return err
	}

	_, body, err := stdlib.UniqueStringMessageContext(ctx, env)
	if err != nil {
		log.Err(err).Msg("could not read envelope")
		return err
	}

	output, err := router.outputJQProgram.RunRaw(body)
	if err != nil {
		return fmt.Errorf(
			"Could not extract output name from payload: %s", err)
	}

	if err := outputs.Forward(output, env.ID()); err != nil {
		log.Err(err).Msg("could not send envelope")
		return fmt.Errorf(
			"could not send envelope: %s", err)
	}

	return nil
}

func init() {
	service.MustRegisterWorker(
		"routing.imperative-router", NewImperativeRouterAsActorService)
}
