package service

import (
	"context"
	"fmt"

	"xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"
)

// ConsumerFunc is a function implementing a consumer
type ConsumerFunc func(context.Context, *api.ActorProcessRequest, Inputs, xbus.Logger) error

// Consumer ...
type Consumer interface {
	Process(context.Context, *api.ActorProcessRequest, Inputs, xbus.Logger) error
}

// ConsumerFactory is a factory of Consumer
type ConsumerFactory func(*api.Actor, xbus.Settings, xbus.Logger) (Consumer, error)

// RegisterConsumer registers a consumer
func RegisterConsumer(name string, consumerFactory ConsumerFactory) (err error) {
	defer func() {
		if r := recover(); r != nil {
			err = fmt.Errorf("%v", r)
		}
	}()
	xbus.RegisterActorService(name, newConsumerServiceFactory(consumerFactory))
	return
}

// MustRegisterConsumer registers a consumer
func MustRegisterConsumer(name string, consumerFactory ConsumerFactory) {
	if err := RegisterConsumer(name, consumerFactory); err != nil {
		panic(err)
	}

}

// RegisterConsumerFunc registers a consumer
func RegisterConsumerFunc(name string, process ConsumerFunc) error {
	return RegisterConsumer(
		name,
		func(*api.Actor, xbus.Settings, xbus.Logger) (Consumer, error) {
			return process, nil
		})
}

// MustRegisterConsumerFunc registers a consumer
func MustRegisterConsumerFunc(name string, process ConsumerFunc) {
	if err := RegisterConsumerFunc(name, process); err != nil {
		panic(err)
	}
}

// UnregisterConsumer unregisters a consumer
func UnregisterConsumer(name string) {
	xbus.UnregisterActorService(name)
}

// Process calls the function
func (fn ConsumerFunc) Process(ctx context.Context, req *api.ActorProcessRequest, inputs Inputs, log xbus.Logger) error {
	return fn(ctx, req, inputs, log)
}

func newConsumerServiceFactory(factory ConsumerFactory) func(*xbus.Actor, xbus.Settings) (xbus.ActorService, error) {
	return func(actor *xbus.Actor, settings xbus.Settings) (xbus.ActorService, error) {
		apiActor := actor.Options.Actor
		consumer, err := factory(apiActor, settings, actor.Log)
		if err != nil {
			return nil, err
		}
		return &actorService{
			actor:   actor,
			service: consumer,
			process: consumerProcess(consumer),
		}, err
	}
}

func consumerProcess(consumer Consumer) func(
	ctx context.Context, svc *actorService, req *api.ActorProcessRequest, send func(*api.ActorProcessingState),
) error {
	return func(ctx context.Context, svc *actorService, req *api.ActorProcessRequest, send func(*api.ActorProcessingState)) error {
		log := xbus.NewLogger(svc.actor.Log.With().Object("processing_context", req.Context).Logger())
		inputs := realInputs{
			ctx:       ctx,
			client:    svc.actor.Client,
			actorID:   svc.actor.ID,
			actorKind: "consumer",
			inputs:    req.Inputs,
		}
		return consumer.Process(ctx, req, inputs, log)
	}
}
